; malloc()

; Every block in the heap, whether allocated or not, consists of four or more
; words.  The LS bit of the first word is 1 if the block is allocated, or 0
; if the block is free.  The LS bit of the last word is always 0.  The 15 MS
; bits of the first and last words contain the number of words in the block,
; including the first and last words.  If the block is allocated, other words
; contain the data.  If the block is free, the second and third words contain
; pointers to the previous and next free blocks, respectively.  A "dummy"
; free block called __endfree is at the beginning and end of the free block
; chain.

BLOCK   STRUC
SIZALL  DW      (?)     ; size and allocation
PREV    DW      (?)     ; previous free block
NEXT    DW      (?)     ; next free block
BLOCK   ENDS

_TEXT   SEGMENT BYTE PUBLIC 'CODE'
_TEXT   ENDS

_DATA           SEGMENT WORD PUBLIC 'DATA'

                PUBLIC  _minimum_block, __endfree

                EXTRN   _heap_top:WORD
                EXTRN   _heap_bottom:WORD

_minimum_block  DW      16
__endfree       BLOCK   <0,0,0>

INIT            EQU     __endfree.SIZALL
FIRSTFREE       EQU     __endfree.NEXT
LASTFREE        EQU     __endfree.PREV

_DATA   ENDS

DGROUP  GROUP   _TEXT, _DATA

_TEXT   SEGMENT
        ASSUME  CS:DGROUP, DS:DGROUP

        PUBLIC  _malloc

_malloc PROC NEAR

        PUSH    SI
        PUSH    DI
        CMP     INIT,0          ; if the heap has not been initialized
        JNZ     MAL2
        PUSH    DS
        POP     ES
        MOV     BX,_heap_top    ; expand memory block to heap_top
        ADD     BX,15           ; round to next higher paragraph
        RCR     BX,1
        MOV     CL,3
        SHR     BX,CL
        MOV     AH,4AH
        INT     21H
        JC      MAL21
        MOV     BX,_heap_top
        MOV     AX,BX
        SUB     AX,_heap_bottom ; number of bytes in block
        MOV     [BX-2],AX       ; to first and last words of block
        MOV     BX,_heap_bottom
        MOV     [BX],AX
        MOV     [BX].PREV,OFFSET DGROUP:__endfree
        MOV     [BX].NEXT,OFFSET DGROUP:__endfree
        MOV     FIRSTFREE,BX
        MOV     LASTFREE,BX
        MOV     INIT,1

MAL2:   MOV     BX,SP           ; add 8 bytes to size, then round
        MOV     DX,[BX+6]       ; upward to next even number
        ADD     DX,8+1
MAL21:  JC      ERROR
        AND     DX,0FFFEH       ; DX = size of block to be allocated
        MOV     SI,FIRSTFREE    ; SI -> beginning of first free block

MAL3:   CMP     SI,OFFSET DGROUP:__endfree
        JZ      ERROR
        MOV     CX,[SI].SIZALL
        SUB     CX,DX           ; CX = excess bytes in block
        JC      MAL5            ; block is not big enough
        MOV     BX,[SI].PREV    ; remove block from free list
        MOV     AX,[SI].NEXT
        MOV     [BX].NEXT,AX
        MOV     BX,[SI].NEXT
        MOV     AX,[SI].PREV
        MOV     [BX].PREV,AX
        CMP     CX,_minimum_block ; is the excess at least the minimum?
        JB      MAL4            ; if not, allocate whole block
        MOV     [SI].SIZALL,DX  ; if so, first shrink this block to DX bytes
        MOV     DI,SI
        ADD     DI,DX           ; DI -> excess portion
        MOV     [DI-2],DX
        MOV     [DI].SIZALL,CX  ; make excess portion CX bytes
        MOV     BX,CX
        MOV     [DI+BX-2],CX
                                ; put the excess into the free chain
        MOV     [DI].PREV,OFFSET DGROUP:__endfree
        MOV     BX,FIRSTFREE    ; BX -> first free block
        MOV     [DI].NEXT,BX
        MOV     FIRSTFREE,DI
        MOV     [BX].PREV,DI

MAL4:   INC     [SI].SIZALL     ; mark block as occupied
        MOV     AX,SI           ; return pointer to second word
        INC     AX
        INC     AX
        JMP     EXIT

MAL5:   MOV     SI,[SI].NEXT    ; next free block
        JMP     MAL3

ERROR:  SUB     AX,AX

EXIT:   POP     DI
        POP     SI
        RET

_malloc ENDP

_TEXT   ENDS
        END
