; table_lookup()

NUMBER_OF_QUEUES  =  16         ; must be a power of 2

_TEXT   SEGMENT BYTE PUBLIC 'CODE'
_TEXT   ENDS

_DATA   SEGMENT WORD PUBLIC 'DATA'

        EXTRN   _first_table_entry:WORD, _case_insensitive_table:BYTE
        EXTRN   _table_entry_size:WORD

NUMBER_OF_ENTRIES  DW      0
LAST_ENTRY         DW      2 DUP (?)
FIRST_IN_QUEUE     DD      NUMBER_OF_QUEUES DUP (0)
BYTES_IN_BUFFER    DW      0
BUFFER             DW      2 DUP (0)

_DATA   ENDS

DGROUP  GROUP   _TEXT, _DATA

TABLE_ENTRY     STRUC
NEXT            DW      2 DUP (?)
ENTRY_NUMBER    DW      (?)
ENTRY_TYPE      DW      (?)
NEXT_IN_QUEUE   DW      2 DUP (?)
TABLE_ENTRY     ENDS


LABEL_NAME       EQU     WORD PTR [BP+8]

; auto variables

NAME_LENGTH      EQU     WORD PTR [BP-2]
PREVIOUS_SEGMENT EQU     WORD PTR [BP-4]
PREVIOUS_OFFSET  EQU     WORD PTR [BP-6]
QUEUE_BASE       EQU     WORD PTR [BP-8]


_TEXT   SEGMENT
        ASSUME  CS:DGROUP, DS:DGROUP

        PUBLIC  _table_lookup
        EXTRN   _insufficient_table_memory:NEAR
        EXTRN   _allocate_permanent:NEAR

_table_lookup PROC NEAR

        PUSH    BP
        PUSH    SI
        PUSH    DI
        MOV     BP,SP
        SUB     SP,8            ; reserve stack space for auto variables

; use a simple hash code to select a queue, and also find the name length

        MOV     BX,LABEL_NAME
        SUB     AL,AL
        MOV     CH,13
HASH1:  MOV     CL,[BX]
        INC     BX
        OR      CL,CL
        JZ      HASH2
        OR      CL,_case_insensitive_table
        MUL     CH
        ADD     AL,CL
        JMP     HASH1
HASH2:  SUB     BX,LABEL_NAME
        MOV     NAME_LENGTH,BX
        AND     AX,NUMBER_OF_QUEUES-1
        ADD     AX,AX
        ADD     AX,AX
        ADD     AX,OFFSET DGROUP:FIRST_IN_QUEUE
        MOV     QUEUE_BASE,AX
          
; look for this entry in the selected queue */

        MOV     BX,AX
        LES     DI,DWORD PTR [BX]       ; ES:DI -> first in queue
        SUB     AX,AX
        MOV     PREVIOUS_OFFSET,AX
        MOV     PREVIOUS_SEGMENT,AX

LOOK1:  MOV     AX,ES
        OR      AX,DI
        JZ      INS1                    ; end of queue and label not found

        MOV     BX,DI
        ADD     BX,_table_entry_size    ; ES:BX -> entry name
        MOV     SI,LABEL_NAME
        MOV     DH,_case_insensitive_table
        CLD
LOOK2:  MOV     DL,ES:[BX]              ; DL = entry name character
        INC     BX
        OR      DL,DH
        LODSB                           ; AL = label name character
        OR      AL,AL                   ; stop comparing at end of label name
        JNZ     LOOK3
        OR      AL,DH
        SUB     AL,DL
        JMP SHORT LOOK4
LOOK3:  OR      AL,DH
        SUB     AL,DL
        JZ      LOOK2                   ; stop comparing at first mismatch

; if it is found, move it to beginning of queue

LOOK4:  JNZ     LOOK6
        MOV     AX,PREVIOUS_OFFSET
        OR      AX,PREVIOUS_SEGMENT
        JNZ     LOOK5
        JMP     EXIT                    ; already at beginning
LOOK5:  MOV     AX,ES:[DI].NEXT_IN_QUEUE ; remove it from queue
        MOV     DX,ES:[DI].NEXT_IN_QUEUE+2 
        PUSH    ES
        LES     BX,DWORD PTR PREVIOUS_OFFSET
        MOV     ES:[BX].NEXT_IN_QUEUE,AX
        MOV     ES:[BX].NEXT_IN_QUEUE+2,DX
        POP     ES
        JMP     INS7                    ; replace it at beginning of queue

; move to next entry in queue

LOOK6:  MOV     PREVIOUS_OFFSET,DI
        MOV     PREVIOUS_SEGMENT,ES
        LES     DI,DWORD PTR ES:[DI].NEXT_IN_QUEUE
        JMP     LOOK1

; if the label was not found, insert it

INS1:   MOV     AX,_table_entry_size    ; create new entry of required size
        ADD     AX,NAME_LENGTH          ; with room for label at end
        PUSH    AX
        CALL    _allocate_permanent
        POP     CX
        MOV     ES,DX
        MOV     DI,AX
        OR      AX,DX
        JNZ     INS2
        CALL    _insufficient_table_memory
INS2:   PUSH    DI
        ADD     DI,_table_entry_size
        MOV     SI,LABEL_NAME
        MOV     CX,NAME_LENGTH
        CLD
        REP     MOVSB                   ; copy label into area at end
        POP     DI
        MOV     AX,_first_table_entry   ; put entry at end of list
        OR      AX,_first_table_entry+2
        JNZ     INS4
        MOV     _first_table_entry,DI
        MOV     _first_table_entry+2,ES
        JMP SHORT INS6
INS4:   MOV     CX,ES
        LES     BX,DWORD PTR LAST_ENTRY
        MOV     ES:[BX].NEXT,DI
        MOV     ES:[BX].NEXT+2,CX
        MOV     ES,CX
INS6:   SUB     AX,AX
        MOV     ES:[DI].NEXT,AX
        MOV     ES:[DI].NEXT+2,AX
        MOV     ES:[DI].ENTRY_TYPE,AX
        MOV     LAST_ENTRY,DI
        MOV     LAST_ENTRY+2,ES
        INC     NUMBER_OF_ENTRIES
        MOV     AX,NUMBER_OF_ENTRIES
        MOV     ES:[DI].ENTRY_NUMBER,AX

; put entry at beginning of queue

INS7:   MOV     BX,QUEUE_BASE
        MOV     AX,[BX]
        MOV     DX,[BX+2]
        MOV     ES:[DI].NEXT_IN_QUEUE,AX
        MOV     ES:[DI].NEXT_IN_QUEUE+2,DX
        MOV     [BX],DI
        MOV     [BX+2],ES

EXIT:   MOV     AX,DI                   ; return pointer in DX:AX
        MOV     DX,ES
        MOV     SP,BP                   ; remove auto variables from stack
        POP     DI
        POP     SI
        POP     BP
        RET

_table_lookup ENDP

_TEXT   ENDS
        END

